@@ -59,11 +59,11 @@ struct _JsonParserPrivate
 
   GError *last_error;
 
-  guint has_assignment : 1;
-  guint is_filename    : 1;
-
   gchar *variable_name;
   gchar *filename;
+
+  guint has_assignment : 1;
+  guint is_filename    : 1;
 };
 
 static const gchar symbol_names[] =
@@ -332,9 +332,80 @@ json_parser_init (JsonParser *parser)
 }
 
 static guint
-json_parse_array (JsonParser *parser,
-                  JsonScanner   *scanner,
-                  gboolean    nested)
+json_parse_value (JsonParser   *parser,
+                  JsonScanner  *scanner,
+                  guint         token,
+                  JsonNode    **node)
+{
+  JsonNode *current_node = parser->priv->current_node;
+  gboolean is_negative = FALSE;
+
+  if (token == '-')
+    {
+      guint next_token = json_scanner_peek_next_token (scanner);
+
+      if (next_token == G_TOKEN_INT ||
+          next_token == G_TOKEN_FLOAT)
+        {
+           is_negative = TRUE;
+           token = json_scanner_get_next_token (scanner);
+        }
+      else
+        return G_TOKEN_INT;
+    }
+
+  switch (token)
+    {
+    case G_TOKEN_INT:
+      *node = json_node_new (JSON_NODE_VALUE);
+      json_node_set_int (*node, is_negative ? scanner->value.v_int64 * -1
+                                            : scanner->value.v_int64);
+      break;
+
+    case G_TOKEN_FLOAT:
+      *node = json_node_new (JSON_NODE_VALUE);
+      json_node_set_double (*node, is_negative ? scanner->value.v_float * -1.0
+                                               : scanner->value.v_float);
+      break;
+
+    case G_TOKEN_STRING:
+      *node = json_node_new (JSON_NODE_VALUE);
+      json_node_set_string (*node, scanner->value.v_string);
+      break;
+
+    case JSON_TOKEN_TRUE:
+    case JSON_TOKEN_FALSE:
+      *node = json_node_new (JSON_NODE_VALUE);
+      json_node_set_boolean (*node, token == JSON_TOKEN_TRUE ? TRUE : FALSE);
+      break;
+
+    case JSON_TOKEN_NULL:
+      *node = json_node_new (JSON_NODE_NULL);
+      break;
+
+    default:
+      {
+        JsonNodeType cur_type;
+
+        *node = NULL;
+
+        cur_type = json_node_get_node_type (current_node);
+        if (cur_type == JSON_NODE_ARRAY)
+          return G_TOKEN_RIGHT_BRACE;
+        else if (cur_type == JSON_NODE_OBJECT)
+          return G_TOKEN_RIGHT_CURLY;
+        else
+          return G_TOKEN_SYMBOL;
+      }
+    }
+
+  return G_TOKEN_NONE;
+}
+
+static guint
+json_parse_array (JsonParser  *parser,
+                  JsonScanner *scanner,
+                  gboolean     nested)
 {
   JsonParserPrivate *priv = parser->priv;
   JsonArray *array;
@@ -356,7 +427,6 @@ json_parse_array (JsonParser *parser,
   while (token != G_TOKEN_RIGHT_BRACE)
     {
       JsonNode *node = NULL;
-      gboolean negative = FALSE;
 
       if (token == G_TOKEN_COMMA)
         {
@@ -387,7 +457,7 @@ json_parse_array (JsonParser *parser,
             }
 
           json_array_add_element (array, node);
-          node->parent = priv->current_node;
+          json_node_set_parent (node, priv->current_node);
 
           g_signal_emit (parser, parser_signals[ARRAY_ELEMENT], 0,
                          array,
@@ -428,7 +498,7 @@ json_parse_array (JsonParser *parser,
             }
 
           json_array_add_element (array, node);
-          node->parent = priv->current_node;
+          json_node_set_parent (node, priv->current_node);
 
           g_signal_emit (parser, parser_signals[ARRAY_ELEMENT], 0,
                          array,
@@ -448,67 +518,20 @@ json_parse_array (JsonParser *parser,
           continue;
         }
 
-      if (token == '-')
+      /* value */
+      token = json_parse_value (parser, scanner, token, &node);
+      if (token != G_TOKEN_NONE || node == NULL)
         {
-          guint next_token = json_scanner_peek_next_token (scanner);
-
-          if (next_token == G_TOKEN_INT ||
-              next_token == G_TOKEN_FLOAT)
-            {
-              negative = TRUE;
-              token = json_scanner_get_next_token (scanner);
-            }
-          else
-            {
-              json_array_unref (array);
-
-              return G_TOKEN_INT;
-            }
-        }
-
-      switch (token)
-        {
-        case G_TOKEN_INT:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_int (node, negative ? scanner->value.v_int64 * -1
-                                            : scanner->value.v_int64);
-          break;
-
-        case G_TOKEN_FLOAT:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_double (node, negative ? scanner->value.v_float * -1.0
-                                               : scanner->value.v_float);
-          break;
-
-        case G_TOKEN_STRING:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_string (node, scanner->value.v_string);
-          break;
-
-        case JSON_TOKEN_TRUE:
-        case JSON_TOKEN_FALSE:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_boolean (node, token == JSON_TOKEN_TRUE ? TRUE : FALSE);
-          break;
-
-        case JSON_TOKEN_NULL:
-          node = json_node_new (JSON_NODE_NULL);
-          break;
-
-        default:
           json_array_unref (array);
-          return G_TOKEN_RIGHT_BRACE;
+          return token;
         }
 
-      if (node)
-        {
-          json_array_add_element (array, node);
-          node->parent = priv->current_node;
+      json_array_add_element (array, node);
+      json_node_set_parent (node, priv->current_node);
 
-          g_signal_emit (parser, parser_signals[ARRAY_ELEMENT], 0,
-                         array,
-                         json_array_get_length (array));
-        }
+      g_signal_emit (parser, parser_signals[ARRAY_ELEMENT], 0,
+                     array,
+                     json_array_get_length (array));
 
       token = json_scanner_get_next_token (scanner);
       if (token != G_TOKEN_COMMA && token != G_TOKEN_RIGHT_BRACE)
@@ -552,7 +575,6 @@ json_parse_object (JsonParser *parser,
     {
       JsonNode *node = NULL;
       gchar *name = NULL;
-      gboolean negative = FALSE;
 
       if (token == G_TOKEN_COMMA)
         {
@@ -612,7 +634,7 @@ json_parse_object (JsonParser *parser,
             }
 
           json_object_set_member (object, name, node);
-          node->parent = priv->current_node;
+          json_node_set_parent (node, priv->current_node);
 
           g_signal_emit (parser, parser_signals[OBJECT_MEMBER], 0,
                          object,
@@ -655,7 +677,7 @@ json_parse_object (JsonParser *parser,
             }
 
           json_object_set_member (object, name, node);
-          node->parent = priv->current_node;
+          json_node_set_parent (node, priv->current_node);
           
           g_signal_emit (parser, parser_signals[OBJECT_MEMBER], 0,
                          object,
@@ -677,68 +699,21 @@ json_parse_object (JsonParser *parser,
           continue;
         }
 
-      if (token == '-')
+      /* value */
+      token = json_parse_value (parser, scanner, token, &node);
+      if (token != G_TOKEN_NONE || node == NULL)
         {
-          guint next_token = json_scanner_peek_next_token (scanner);
-
-          if (next_token == G_TOKEN_INT || next_token == G_TOKEN_FLOAT)
-            {
-              negative = TRUE;
-              token = json_scanner_get_next_token (scanner);
-            }
-          else
-            {
-              g_free (name);
-              json_object_unref (object);
-
-              return G_TOKEN_INT;
-            }
-        }
-
-      switch (token)
-        {
-        case G_TOKEN_INT:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_int (node, negative ? scanner->value.v_int64 * -1
-                                            : scanner->value.v_int64);
-          break;
-
-        case G_TOKEN_FLOAT:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_double (node, negative ? scanner->value.v_float * -1.0
-                                               : scanner->value.v_float);
-          break;
-
-        case G_TOKEN_STRING:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_string (node, scanner->value.v_string);
-          break;
-
-        case JSON_TOKEN_TRUE:
-        case JSON_TOKEN_FALSE:
-          node = json_node_new (JSON_NODE_VALUE);
-          json_node_set_boolean (node, token == JSON_TOKEN_TRUE ? TRUE : FALSE);
-          break;
-
-        case JSON_TOKEN_NULL:
-          node = json_node_new (JSON_NODE_NULL);
-          break;
-
-        default:
-          g_free (name);
           json_object_unref (object);
-          return G_TOKEN_SYMBOL;
+          g_free (name);
+          return token;
         }
 
-      if (node)
-        {
-          json_object_set_member (object, name, node);
-          node->parent = priv->current_node;
-          
-          g_signal_emit (parser, parser_signals[OBJECT_MEMBER], 0,
-                         object,
-                         name);
-        }
+      json_object_set_member (object, name, node);
+      json_node_set_parent (node, priv->current_node);
+
+      g_signal_emit (parser, parser_signals[OBJECT_MEMBER], 0,
+                     object,
+                     name);
 
       g_free (name);
 
@@ -884,8 +859,8 @@ json_parse_statement (JsonParser  *parser,
 
 static void
 json_scanner_msg_handler (JsonScanner *scanner,
-                          gchar    *message,
-                          gboolean  is_error)
+                          gchar       *message,
+                          gboolean     is_error)
 {
   JsonParser *parser = scanner->user_data;
   JsonParserPrivate *priv = parser->priv;
@@ -1130,6 +1105,7 @@ json_parser_load_from_data (JsonParser   *parser,
 
   priv->is_filename = FALSE;
   g_free (priv->filename);
+  priv->filename = NULL;
 
   internal_error = NULL;
   if (!json_parser_load (parser, data, length, &internal_error))
@@ -1208,7 +1184,8 @@ json_parser_get_current_pos (JsonParser *parser)
 /**
  * json_parser_has_assignment:
  * @parser: a #JsonParser
- * @variable_name: return location for the variable name, or %NULL
+ * @variable_name: (out) (allow-none): Return location for the variable
+ *   name, or %NULL
  *
  * A JSON data stream might sometimes contain an assignment, like:
  *
